@@ -7,7 +7,9 @@ use Config;
 
 use 5.022; # for $Config{longdblkind}
 
-use constant NV_IS_DOUBLE       => $Config{nvsize} == 8 ? 1 : 0;
+use constant NAN_COMPARE_BUG    => $Math::MPFR::VERSION < 4.23 ? 1 : 0;
+
+use constant NV_IS_DOUBLE       => $Config{nvsize} == 8        ? 1 : 0;
 
 use constant NV_IS_DOUBLEDOUBLE => $Config{nvsize} != 8 &&
                                    ($Config{longdblkind} >=5 && $Config{longdblkind} <= 8) ? 1 : 0;
@@ -56,14 +58,16 @@ require Exporter;
 @Math::FakeDD::EXPORT_OK = qw(
   NV_IS_DOUBLE NV_IS_DOUBLEDOUBLE NV_IS_QUAD NV_IS_80BIT_LD
   dd_abs dd_add dd_add_eq dd_assign dd_atan2 dd_cmp dd_cos dd_dec dd_div dd_div_eq dd_eq dd_exp
-  dd_gt dd_gte dd_int dd_log dd_lt dd_lte
-  dd_mul dd_mul_eq dd_neq dd_pow dd_pow_eq dd_repro dd_sin dd_spaceship dd_sqrt dd_stringify
+  dd_gt dd_gte dd_hex dd_inf dd_is_inf dd_is_nan dd_int dd_log dd_lt dd_lte
+  dd_mul dd_mul_eq dd_nan dd_neq dd_pow dd_pow_eq dd_repro dd_sin dd_spaceship dd_sqrt dd_stringify
   dd_sub dd_sub_eq
+  dd2mpfr mpfr2dd mpfr_any_prec2dd mpfr2098
+  printx sprintx unpackx
 );
 
 %Math::FakeDD::EXPORT_TAGS = (all =>[@Math::FakeDD::EXPORT_OK]);
 
-$Math::FakeDD::VERSION =  0.01;
+$Math::FakeDD::VERSION =  '0.02';
 
 sub new {
 
@@ -136,6 +140,34 @@ sub dd_add {
 
 }
 
+sub dd_add_4196 {
+  # Used only in testing.
+  die "Wrong number of arguments given to dd_add()"
+    if @_ > 2;
+
+  my ($rop1, $rop2);
+
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    $rop1 = dd2mpfr(shift);
+  }
+  else {
+    $rop1 = mpfr2098(shift);
+  }
+
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    $rop2 = dd2mpfr(shift);
+  }
+  else {
+    $rop2 = mpfr2098(shift);
+  }
+
+  my $ret = Rmpfr_init2(4196);
+
+  Rmpfr_add($ret, $rop1, $rop2, MPFR_RNDN);
+  return mpfr_any_prec2dd($ret);
+
+}
+
 sub dd_add_eq {
   # When dd_add_eq is called via overloading of '+=' a
   # third argument (which we CAN ignore) will be provided
@@ -243,6 +275,7 @@ sub dd_cmp {
   }
 
   return $rop1 <=> $rop2; # "<=>" is "Math::MPFR::overload_spaceship"
+                          # and will return undef if a NaN is involved.
 }
 
 sub dd_cos {
@@ -263,7 +296,18 @@ sub dd_cos {
 sub dd_dec {
   die "Wrong arg given to dd_dec()"
     unless ref($_[0]) eq 'Math::FakeDD';
-  return decimalize(dd2mpfr(shift));
+  my $mpfr = dd2mpfr(shift);
+
+  if(!Rmpfr_regular_p($mpfr)) {
+    return '0.0'   if Rmpfr_zero_p($mpfr);
+    return 'NaN' if Rmpfr_nan_p($mpfr);
+
+    # must be an inf
+    return 'Inf' if $mpfr > 0;
+    return '-Inf';
+  }
+
+  return decimalize($mpfr);
 }
 
 sub dd_div {
@@ -297,6 +341,34 @@ sub dd_div {
 
 }
 
+sub dd_div_4196 {
+  # Used only in testing.
+  die "Wrong number of arguments given to dd_div()"
+    if @_ > 2;
+
+  my ($rop1, $rop2);
+
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    $rop1 = dd2mpfr(shift);
+  }
+  else {
+    $rop1 = mpfr2098(shift);
+  }
+
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    $rop2 = dd2mpfr(shift);
+  }
+  else {
+    $rop2 = mpfr2098(shift);
+  }
+
+  my $ret = Rmpfr_init2(4196);
+
+  Rmpfr_div($ret, $rop1, $rop2, MPFR_RNDN);
+  return mpfr_any_prec2dd($ret);
+
+}
+
 sub dd_div_eq {
   # When dd_div_eq is called via overloading of '/=' a
   # third argument (which we CAN ignore) will be provided
@@ -336,8 +408,9 @@ sub dd_eq {
   # third argument (which we CAN ignore) will be provided
   die "Wrong number of args passed to dd_eq()"
     if(@_ > 3);
-  return 0 if dd_cmp(shift, shift); # unequal
-  return 1;                         # equal
+  my $cmp = dd_cmp(shift, shift);
+  return 0 if $cmp || !defined $cmp; # not equal
+  return 1;                          # equal
 }
 
 sub dd_exp {
@@ -385,6 +458,27 @@ sub dd_gte {
   return 1;                                           # greater than or equal
 }
 
+sub dd_hex {
+  die "Wrong arg given to dd_dec()"
+    unless ref($_[0]) eq 'Math::FakeDD';
+
+  my $mpfr = dd2mpfr(shift);
+
+  if(!Rmpfr_regular_p($mpfr)) {
+    return '0x0p+0'   if Rmpfr_zero_p($mpfr);
+    return 'NaN' if Rmpfr_nan_p($mpfr);
+
+    # must be an inf
+    return 'Inf' if $mpfr > 0;
+    return '-Inf';
+  }
+
+  my $buffer;
+  Rmpfr_sprintf($buffer, "%Ra", $mpfr, 528);
+
+  return $buffer;
+}
+
 sub dd_int {
   # Don't fall for the idea that we can just do int(msd), int(lsd)
   # when $_[0] is a Math::FakeDD object. I tried that and it doesn't
@@ -477,6 +571,34 @@ sub dd_mul {
 
 }
 
+sub dd_mul_4196 {
+  # Used only in testing.
+  die "Wrong number of arguments given to dd_mul()"
+    if @_ > 2;
+
+  my ($rop1, $rop2);
+
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    $rop1 = dd2mpfr(shift);
+  }
+  else {
+    $rop1 = mpfr2098(shift);
+  }
+
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    $rop2 = dd2mpfr(shift);
+  }
+  else {
+    $rop2 = mpfr2098(shift);
+  }
+
+  my $ret = Rmpfr_init2(4196);
+
+  Rmpfr_mul($ret, $rop1, $rop2, MPFR_RNDN);
+  return mpfr_any_prec2dd($ret);
+
+}
+
 sub dd_mul_eq {
   # When dd_mul_eq is called via overloading of '*=' a
   # third argument (which we CAN ignore) will be provided
@@ -516,8 +638,9 @@ sub dd_neq {
   # third argument (which we CAN ignore) will be provided
   die "Wrong number of args passed to dd_neq()"
     if(@_ > 3);
-  return 1 if dd_cmp(shift, shift); # unequal
-  return 0;                         # equal
+  my $cmp = dd_cmp(shift, shift);
+  return 1 if $cmp || !defined $cmp; # not equal
+  return 0;                          # equal
 }
 
 sub dd_pow {
@@ -587,7 +710,70 @@ sub dd_pow_eq {
 sub dd_repro {
   die "Arg given to dd_repro() must be a Math::FakeDD object"
     unless ref($_[0]) eq 'Math::FakeDD';
-  return mpfrtoa(dd2mpfr(shift));
+
+  my $arg = shift;
+
+  return 'NaN' if dd_is_nan($arg);
+
+  if(dd_is_inf($arg)) {
+    return'Inf' if $arg > 0;
+    return'-Inf';
+  }
+
+  return '0.0' if $arg == 0;
+
+  my $neg = 0;
+  my $mpfr = dd2mpfr($arg);
+  if($mpfr < 0) {
+    Rmpfr_neg($mpfr, $mpfr, MPFR_RNDN);
+    $neg = 1;
+  }
+  my @v = Rmpfr_deref2($mpfr, 2, 0, MPFR_RNDN);
+
+  if($arg->{lsd} == 0) {
+    if(abs($arg->{msd}) >= 2 ** -1022) {
+      # msd is NOT subnormal, so use 53-bit precision
+      Rmpfr_prec_round($mpfr, 53, MPFR_RNDN)
+        if Rmpfr_get_exp($mpfr) <= 0;
+    }
+    else {
+      # msd is subnormal; the exponent, with the mpfr
+      # library, will be in the range (-1022 .. -1073)
+      # Rmpfr_prec_round($mpfr, $v[1] + 1074, MPFR_RNDN);
+      Rmpfr_prec_round($mpfr, Rmpfr_get_exp($mpfr) + 1074, MPFR_RNDN);
+    }
+  }
+  else {
+    $v[0] =~ s/0+$//; # remove all trailing zeroes
+    my $m_lsd = Rmpfr_init2(53);
+    Rmpfr_set_d($m_lsd, $arg->{lsd}, MPFR_RNDN);
+
+    my @v_lsd = Rmpfr_deref2($m_lsd, 2, 0, MPFR_RNDN);
+    $v_lsd[0] =~ s/0+$//; # remove all trailing zeroes
+    my $v_lsd_len = length($v_lsd[0]);
+    my $correction = 0;
+
+    # $correction will be altered to the pertinent +ve
+    # value (below) when length($v[0]) is understating
+    # the required precision.
+
+    if(abs($arg->{lsd}) >= 2 ** -1022) {
+      # lsd is NOT subnormal
+      $correction = 53 - $v_lsd_len
+         if $v_lsd_len < 53;
+    }
+    else {
+      # lsd is subnormal
+      $correction = $v_lsd[1] + 1074 - $v_lsd_len
+        if $v_lsd_len < $v_lsd[1] + 1074;
+
+    }
+
+    Rmpfr_prec_round($mpfr, length($v[0]) + $correction, MPFR_RNDN);
+  }
+
+  return '-' . mpfrtoa($mpfr) if $neg;
+  return mpfrtoa($mpfr);
 }
 
 sub dd_sin {
@@ -616,7 +802,17 @@ sub dd_spaceship {
   $correction = -1
     if (@_ == 3 && $_[2]);
 
-  return $correction * dd_cmp(shift, shift);
+  my($arg1, $arg2) = (shift, shift);
+
+  if(NAN_COMPARE_BUG) { # Fixed in Math-MPFR-4.23 and later.
+    if(dd_is_nan(Math::FakeDD->new($arg1)) || dd_is_nan(Math::FakeDD->new($arg2))) {
+      return undef;
+    }
+  }
+
+  my $cmp = dd_cmp($arg1, $arg2);
+  return $correction * $cmp if defined $cmp;
+  return $cmp;
 }
 
 sub dd_sqrt {
@@ -677,6 +873,33 @@ sub dd_sub {
   return mpfr2dd($rop1);
 }
 
+sub dd_sub_4196 {
+  # Used only in testing.
+  die "Wrong number of arguments given to dd_sub()"
+    if @_ > 2;
+
+  my ($rop1, $rop2);
+
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    $rop1 = dd2mpfr(shift);
+  }
+  else {
+    $rop1 = mpfr2098(shift);
+  }
+
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    $rop2 = dd2mpfr(shift);
+  }
+  else {
+    $rop2 = mpfr2098(shift);
+  }
+
+  my $ret = Rmpfr_init2(4196);
+
+  Rmpfr_sub($ret, $rop1, $rop2, MPFR_RNDN);
+  return mpfr_any_prec2dd($ret);
+}
+
 sub dd_sub_eq {
   # When dd_sub_eq is called via overloading of '-=' a
   # third argument (which we CAN ignore) will be provided
@@ -740,7 +963,35 @@ sub mpfr2dd {
     $h{lsd} = 0;
     return bless(\%h);
   }
-  $mpfr -= $msd;
+  Rmpfr_sub_d($mpfr, $mpfr, $msd, MPFR_RNDN);
+  my $lsd = Rmpfr_get_d($mpfr, MPFR_RNDN);
+  $h{msd} = $msd;
+  $h{lsd} = $lsd;
+  return bless(\%h);
+}
+
+sub mpfr_any_prec2dd {
+  # Converts a Math::MPFR object of any
+  # precision to a Math::FakeDD object.
+  my %h;
+
+  die "Arg given to mpfr2dd() must be a Math::MPFR object"
+    unless ref($_[0]) eq 'Math::MPFR';
+
+  my $prec_in = Rmpfr_get_prec($_[0]);
+  my $mpfr_prec = $prec_in <= 2098 ? 2098 : $prec_in;
+
+  my $mpfr = Rmpfr_init2($mpfr_prec);
+  Rmpfr_set($mpfr, shift, MPFR_RNDN);
+
+  my $msd = Rmpfr_get_d($mpfr, MPFR_RNDN);
+  if($msd == 0 || $msd != $msd || $msd / $msd != 1) { # $msd is zero, nan, or inf.
+    $h{msd} = $msd;
+    $h{lsd} = 0;
+    return bless(\%h);
+  }
+
+  Rmpfr_sub_d($mpfr, $mpfr, $msd, MPFR_RNDN);
   my $lsd = Rmpfr_get_d($mpfr, MPFR_RNDN);
   $h{msd} = $msd;
   $h{lsd} = $lsd;
@@ -755,7 +1006,7 @@ sub mpfr2098 {
   # Arg must be one of PV (string), IV (integer), UV (unsigned integer),
   # NV (perl floating point type) or Math::MPFR object.
 
-  die "Invalid arg passed internally to mpfr2098()"
+  die "Invalid arg ($itsa) passed internally to mpfr2098()"
     unless ($itsa > 0 && $itsa <= 4);
 
   my $arg = shift;
@@ -830,6 +1081,53 @@ sub dd_false {
   return 0;
 }
 
+sub dd_is_inf {
+  die "Not a Math::FakeDD object passed to dd_is_inf()"
+    unless ref($_[0]) eq 'Math::FakeDD';
+
+  return 1 if Rmpfr_inf_p(Math::MPFR->new($_[0]->{msd}));
+  return 0;
+}
+
+sub dd_is_nan {
+  die "Not a Math::FakeDD object passed to dd_is_nan()"
+    unless ref($_[0]) eq 'Math::FakeDD';
+
+  return 1 if Rmpfr_nan_p(Math::MPFR->new($_[0]->{msd}));
+  return 0;
+}
+
+sub dd_inf {
+  my $inf = Math::MPFR->new();
+  Rmpfr_set_inf($inf, defined($_[0]) ?  shift : 0); # Will be -Inf only if $_[0] < 0
+  my %h = (msd => Rmpfr_get_d($inf, MPFR_RNDN), lsd => 0.0);
+  return bless \%h;
+}
+
+sub dd_nan {
+  my %h = (msd => Rmpfr_get_d(Math::MPFR->new(), MPFR_RNDN), lsd => 0.0);
+  return bless \%h;
+}
+
+sub printx {
+  print sprintx(shift);
+}
+
+sub sprintx {
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    my $self = shift;
+    return "[" . sprintf("%a", $self->{msd}) . " " . sprintf("%a", $self->{lsd}) . "]";
+  }
+  die "Wrong arg given to sprintx()";
+}
+
+sub unpackx {
+  if(ref($_[0]) eq 'Math::FakeDD') {
+    my $self = shift;
+    return "[" . unpack("H*", pack("d>", $self->{msd})) . " " . unpack("H*", pack("d>", $self->{lsd})) . "]";
+  }
+  die "Wrong arg given to unpackx()";
+}
 
 1;
 
